'use strict';

function removeFromTo(array, from, to) {
    array.splice(from,
        !to ||
        1 + to - from + (!(to < 0 ^ from >= 0) && (to < 0 || -1) * array.length));
    return array.length;
}

/**
 * История действий, которые можно оменить или повторить
 */
class ActionHistory {
    /**
     * Создает пустую историю
     */
    constructor() {
        this.actions = [];
        this.currentIndex = -1;
        this.actionLimit = 0;
        this.isExecuting = false;
    }

    /**
     * Повторяет или отменяет действие из истории
     * @param {EditAction} action - действие из этой истории
     * @param {string} command - undo или redo
     * @returns {ActionHistory}
     */
    execute(action, command) {
        if (!action || typeof action[command] !== "function") {
            return false;
        }
        this.isExecuting = true;

        let succeeded = action[command]();

        this.isExecuting = false;
        return succeeded;
    }

    /**
     * Добавить действие в историю. Удалит все дейтсвия после добавленного (если были)
     * и удалит самые старые дейтсвия, если задан максимальный размер истории
     * @param {EditAction} action - действие для добавления
     * @returns {ActionHistory}
     */
    add(action) {
        if (this.isExecuting) {
            return this;
        }
        // удалить все действия после этого, если есть
        this.actions.splice(this.currentIndex + 1, this.actions.length - this.currentIndex);

        this.actions.push(action)

        // Если задано максимальное число действий, удалить лишние из начала истории
        if (this.limit && this.actions.length > this.actionLimit) {
            removeFromTo(this.actions, 0, -(this.actionLimit+1));
        }

        // set the current index to the end
        this.currentIndex = this.actions.length - 1;
        return this;
    }

    /**
     * Создать новое действие и добавить в историю
     * @param {string} cellId - Id элемента dom в котором осуществлено редактирование.
     * @param {function} actionDo - Код, выполнение которого повторит действие
     * @param {function} actionUndo - Код, выполнение которого отменит результат действия
     * @return {EditAction} - добавленное действие
     */
    addAction(cellId, actionDo, actionUndo) {
        let action = new EditAction(cellId, actionDo, actionUndo, this);
        this.add(action);
        return action;
    }

    /**
     * Отменить текущее действие
     * @returns {ActionHistory}
     */
    undo() {
        var action = this.actions[this.currentIndex];
        if (!action) {
            return false;
        }
        let couldUndo = this.execute(action, "undo");
        if (couldUndo) {
            action.deactivate();
            this.currentIndex -= 1;
        }
        return couldUndo;
    }

    /**
     * Повторить текущее действие
     * @returns {ActionHistory}
     */
    redo() {
        var action = this.actions[this.currentIndex + 1];
        if (!action) {
            return false;
        }
        let couldRedo = this.execute(action, "redo");
        if (couldRedo) {
            action.deactivate();
            this.currentIndex += 1;
        }
        return couldRedo;
    }

    /**
     * Очистить историю
     */
    clear() {
        this.actions = [];
        this.currentIndex = -1;
    }

    /**
     * Есть ли действия для отмены
     * @returns {boolean}
     */
    hasUndo() {
        return this.currentIndex !== -1;
    }

    /**
     * Есть ли действия для повтора
     * @returns {boolean}
     */
    hasRedo() {
        return this.currentIndex < (this.actions.length - 1);
    }

    /**
     * Вернуть стек действий
     * @returns {Array}
     */
    getActions() {
        return this.actions;
    }

    /**
     * Вернуть номер текущего действия
     * @returns {*|number}
     */
    getIndex() {
        return this.currentIndex;
    }

    /**
     * Установить максимальный размер истории
     * @param lim
     */
    setLimit(lim) {
        this.actionLimit = lim;
    }

    /**
     * Удалить действие из истории
     * @param action
     */
    removeAction(action) {
        let actionIndex = this.actions.indexOf(action);
        if (actionIndex === this.currentIndex) {
            this.currentIndex -= 1;
        }
        this.actions.splice(actionIndex, 1);
        if (this.actions.length = 0) {
            this.currentIndex = -1;
        }
    }
}

/**
 * Представляет действие по редактированию, которое отправляется на сервер
 * и которое можно отменить или повторить
 */
class EditAction {

    /**
     * Создает неактивное действие
     * @param {string} cellId - Id элемента dom в котором осуществлено редактирование.
     * @param {function} actionDo - Код, выполнение которого повторит действие
     * @param {function} actionUndo - Код, выполнение которого отменит результат действия
     * @param {ActionHistory} parentHistory - История, в которую входит данное действие
     */
    constructor(cellId, actionDo, actionUndo, parentHistory) {
        this.cellId = cellId;
        this.actionDo = actionDo;
        this.actionUndo = actionUndo;
        this.active = false;
        this.history = parentHistory;
    }

    /**
     * Поторить действие
     * @return {boolean} Удалось ли повторить последнее действие.
     * Не удается только если оно еще не подтверждено.
     */
    redo() {
        if (this.active) {
            this.actionDo(this);
            return true;
        } else {
            return false;
        }
    }

    /**
     * Отменить действие
     * @returns {boolean} Удалось ли отменить последнее действие.
     * Не удается только если оно еще не подтверждено.
     */
    undo() {
        if (this.active) {
            this.actionUndo(this);
            return true;
        } else {
            return false;
        }
    }

    /**
     * Сделать действие активным, то есть доступным для отмены или повтора
     */
    activate() {
        this.active = true;
    }

    /**
     * Сделать действие неактивным, то есть недоступным для отмены или повтора
     */
    deactivate() {
        this.active = false;
    }

    /**
     * Удалить это действие
     */
    remove() {
        this.history.removeAction(this);
    }

}

if (typeof define === 'function' && typeof define.amd === 'object' && define.amd) {
    // AMD. Register as an anonymous module.
    define(function() {
        return ActionHistory;
    });
} else if (typeof module !== 'undefined' && module.exports) {
    module.exports = ActionHistory;
} else {
    window.ActionHistory = ActionHistory;
}